from topax.sdfs import (
    sphere,
    union,
    translate,
    intersect,
    subtract,
    scale
)

def make_part():
    return scale(subtract(sphere(0.5, [0.4, 0, 0]), sphere(0.5, x=-0.4)), 1.5)



import numpy as np
import OpenGL.GL as gl

def norm(v, axis=-1, keepdims=False, eps=0.0):
    return np.sqrt((v*v).sum(axis, keepdims=keepdims).clip(eps))

def normalize(v, axis=-1, eps=1e-20):
    return v/norm(v, axis, keepdims=True, eps=eps)

def rotation_matrix(angle, axis):
    s, c = np.sin(angle), np.cos(angle)
    match axis:
        case 'x':
            return np.array([[1., 0., 0.], [0., c, -s], [0., s, c]])
        case 'y':
            return np.array([[c, 0., s], [0., 1., 0.], [-s, 0., c]])
        case 'z':
            return np.array([[c, -s, 0.], [s, c, 0.], [0., 0., 1.]])
        case _:
            raise ValueError(f"Axis must be 'x' 'y' or 'z', not {axis}")
        
def rotation_matrix_about_vector(angle, axis_vec):
    axis_vec = np.asarray(axis_vec, dtype=float)
    axis_vec = axis_vec / np.linalg.norm(axis_vec)
    x, y, z = axis_vec
    c = np.cos(angle)
    s = np.sin(angle)
    C = 1 - c
    return np.array([
        [c + x*x*C,     x*y*C - z*s, x*z*C + y*s],
        [y*x*C + z*s,   c + y*y*C,   y*z*C - x*s],
        [z*x*C - y*s,   z*y*C + x*s, c + z*z*C]
    ])

def compile_shader(src, shader_type):
    shader = gl.glCreateShader(shader_type)
    gl.glShaderSource(shader, src)
    gl.glCompileShader(shader)
    if not gl.glGetShaderiv(shader, gl.GL_COMPILE_STATUS):
        raise RuntimeError(gl.glGetShaderInfoLog(shader).decode())
    return shader

def resolve_type(item):
    if isinstance(item, float):
        return 'float'
    elif hasattr(item, '__iter__'):
        match len(item):
            case 3:
                return 'vec3'
            case 2:
                return 'vec2'
            case 1:
                return 'float'
            case _:
                raise NotImplementedError(f"cannot resolve type of value {item}")
    else:
        raise NotImplementedError(f"cannot resolve type of value {item}")
    
def make_shader_const(item, type):
    match type:
        case 'vec3':
            assert len(item) == 3, f"length of vec3 must be 3, but got {len(item)} for item {item}"
            return f"vec3({",".join(item)})"
        case 'vec2':
            assert len(item) == 2
            return f"vec2({",".join(item)})"
        case 'float':
            if hasattr(item, '__iter__'):
                assert len(item) == 1
                item = item[0]
            return f"{item}"
        case _: raise NotImplementedError(f"cannot make shader constant from type {type}")



from enum import Enum
from typing import Any
from dataclasses import dataclass

from topax._utils import resolve_type

# TODO: make this enum numbering better
class OpType(Enum):
    ADD = 1
    SUB = 2
    MUL = 3
    DIV = 4
    LEN = 5
    NORM = 6
    SQRT = 7
    SIN = 8
    COS = 9
    MIN = 10
    MAX = 11
    X = 12
    Y = 13
    Z = 14
    XY = 15
    XZ = 16
    YZ = 17
    NEG = 18
    ABS = 19

@dataclass(frozen=True)
class Op:
    opcode: OpType
    lhs: Any
    rhs: Any = None
    rettype: str = ""

    # TODO: improve this logic to be more robust
    def _set_rettype(self, rettype=None):
        if rettype is not None:
            object.__setattr__(self, 'rettype', rettype)
        else:
            lhs_rettype = self.lhs.rettype
            rhs_rettype = self.rhs.rettype if self.rhs is not None else ''
            if lhs_rettype == 'vec3' or rhs_rettype == 'vec3':
                object.__setattr__(self, 'rettype', 'vec3')
            elif lhs_rettype == 'vec2' or rhs_rettype == 'vec2':
                object.__setattr__(self, 'rettype', 'vec2')
            else:
                object.__setattr__(self, 'rettype', 'float')

    def __post_init__(self):
        if not isinstance(self.lhs, Op) and not isinstance(self.lhs, Const):
            object.__setattr__(self, 'lhs', Const(None, self.lhs))
        if self.rhs is not None and not isinstance(self.rhs, Op) and not isinstance(self.rhs, Const):
            object.__setattr__(self, 'rhs', Const(None, self.rhs))
        if self.rettype == "":
            match self.opcode:
                case OpType.ADD: self._set_rettype()
                case OpType.SUB: self._set_rettype()
                case OpType.MUL: self._set_rettype()
                case OpType.DIV: self._set_rettype()
                case OpType.LEN: self._set_rettype('float')
                case OpType.NORM: self._set_rettype()
                case OpType.SQRT: self._set_rettype()
                case OpType.SIN: self._set_rettype()
                case OpType.COS: self._set_rettype()
                case OpType.MIN: self._set_rettype()
                case OpType.MAX: self._set_rettype()
                case OpType.X: self._set_rettype('float')
                case OpType.Y: self._set_rettype('float')
                case OpType.Z: self._set_rettype('float')
                case OpType.XY: self._set_rettype('vec2')
                case OpType.XZ: self._set_rettype('vec2')
                case OpType.YZ: self._set_rettype('vec2')
                case OpType.NEG: self._set_rettype(self.lhs.rettype)
                case OpType.ABS: self._set_rettype(self.lhs.rettype)
                case _: raise NotImplementedError(f"rettype for opcode {self.opcode} not supported")

    @property
    def x(self): return Op(OpType.X, self)
    @property
    def y(self): return Op(OpType.Y, self)
    @property
    def z(self): return Op(OpType.Z, self)
    @property
    def xy(self): return Op(OpType.XY, self)
    @property
    def xz(self): return Op(OpType.XZ, self)
    @property
    def yz(self): return Op(OpType.YZ, self)

    def __add__(self, rhs): return Op(OpType.ADD, self, rhs)
    def __radd__(self, lhs): return Op(OpType.ADD, lhs, self)
    
    def __sub__(self, rhs): return Op(OpType.SUB, self, rhs)
    def __rsub__(self, lhs): return Op(OpType.SUB, lhs, self)

    def __pos__(self): return self
    def __neg__(self): return Op(OpType.NEG, self)
    
    def __mul__(self, rhs): return Op(OpType.MUL, self, rhs)
    def __rmul__(self, lhs): return Op(OpType.MUL, lhs, self)

    def __truediv__(self, rhs): return Op(OpType.DIV, self, rhs)
    def __rtruediv__(self, lhs): return Op(OpType.DIV, lhs, self)
    
    def __repr__(self):
        if self.rhs is not None:
            return f"{self.opcode}({self.lhs},{self.rhs})->{self.rettype}"
        else:
            return f"{self.opcode}({self.lhs})->{self.rettype}"

# TODO: maybe get rid of this and just make it an OpType?    
@dataclass(frozen=True)
class Const:
    sdf: Any
    param: str | Any
    rettype: str | None = None

    def __post_init__(self):
        if self.rettype is None:
            type = resolve_type(self.param)
            object.__setattr__(self, 'rettype', type)

    def resolve_value(self):
        assert self.sdf is not None
        return self.sdf[self.param]
    
    @property
    def x(self): return Op(OpType.X, self)
    @property
    def y(self): return Op(OpType.Y, self)
    @property
    def z(self): return Op(OpType.Z, self)
    @property
    def xy(self): return Op(OpType.XY, self)
    @property
    def xz(self): return Op(OpType.XZ, self)
    @property
    def yz(self): return Op(OpType.YZ, self)

    def __add__(self, rhs): return Op(OpType.ADD, self, rhs)
    def __radd__(self, lhs): return Op(OpType.ADD, lhs, self)
    
    def __sub__(self, rhs): return Op(OpType.SUB, self, rhs)
    def __rsub__(self, lhs): return Op(OpType.SUB, lhs, self)
    
    def __pos__(self): return self
    def __neg__(self): return Op(OpType.NEG, self)

    def __mul__(self, rhs): return Op(OpType.MUL, self, rhs)
    def __rmul__(self, lhs): return Op(OpType.MUL, lhs, self)

    def __truediv__(self, rhs): return Op(OpType.DIV, self, rhs)
    def __rtruediv__(self, lhs): return Op(OpType.DIV, lhs, self)

    def __repr__(self): return f"Const({type(self.sdf).__name__};{self.param};{self.rettype})"

    def __eq__(self, other): return self.sdf == other.sdf and self.param == other.param



import abc

from numpy.typing import ArrayLike

from topax._utils import resolve_type
from topax.ops import Op, OpType, Const

class SDF(abc.ABC):
    def __init__(self, *args, **kwargs):
        assert len(args) == 0
        self._values = {}
        for k in kwargs:
            item = kwargs[k]
            self._values[k] = item
            v = Const(self, k, resolve_type(item))
            setattr(self, k, v)

    @abc.abstractmethod
    def sdf_definition(self, p) -> Op | Const:
        raise NotImplementedError()

    def __call__(self, p):
        return self.sdf_definition(p)
    
    def __getitem__(self, key):
        return self._values[key]

class empty(SDF):
    def __init__(self):
        super().__init__()

    def sdf_definition(self, p):
        return Const(None, 'uintBitsToFloat(0x7F800000u)', 'float')

class sphere(SDF):
    def __init__(
        self,
        radius: float, 
        center: ArrayLike | None=None, 
        x: float | None=None,
        y: float | None=None,
        z: float | None=None
    ):
        if center is not None:
            super().__init__(radius=radius, center=center)
        else:
            center = [x, y, z]
            if all(e is None for e in center):
                super().__init__(radius=radius)
            else:
                center = [0.0 if e is None else e for e in center]
                super().__init__(radius=radius, center=center)

    def sdf_definition(self, p):
        if hasattr(self, 'center'):
            return Op(OpType.LEN, p - self.center) - self.radius
        else:
            return Op(OpType.LEN, p) - self.radius
        
class box(SDF):
    def __init__(
        self,
        size: ArrayLike,
    ):
        super().__init__(size=size)

    def sdf_definition(self, p):
        q = Op(OpType.ABS, p) - self.size
        return Op(OpType.LEN, Op(OpType.MAX, q, 0.0)) + Op(OpType.MIN, Op(OpType.MAX, q.x, Op(OpType.MAX, q.y, q.z)), 0.0)
    
class translate(SDF):
    def __init__(self, sdf: SDF, offset: ArrayLike):
        self.sdf = sdf
        super().__init__(offset=offset)

    def sdf_definition(self, p):
        return self.sdf(p - self.offset)
    
class union(SDF):
    def __init__(self, *sdfs: SDF):
        self.sdfs = sdfs
        super().__init__()

    def sdf_definition(self, p):
        if len(self.sdfs) == 1:
            return self.sdfs[0](p)
        oper = Op(OpType.MIN, self.sdfs[0](p), self.sdfs[1](p))
        for i in range(2, len(self.sdfs)):
            oper = Op(OpType.MIN, oper, self.sdfs[i](p))
        return oper
    
class intersect(SDF):
    def __init__(self, *sdfs: SDF):
        self.sdfs = sdfs
        super().__init__()

    def sdf_definition(self, p):
        if len(self.sdfs) == 1:
            return self.sdfs[0](p)
        oper = Op(OpType.MAX, self.sdfs[0](p), self.sdfs[1](p))
        for i in range(2, len(self.sdfs)):
            oper = Op(OpType.MAX, oper, self.sdfs[i](p))
        return oper
    
class subtract(SDF):
    def __init__(self, sdf: SDF, tool: SDF):
        self.sdf = sdf
        self.tool = tool
        super().__init__()

    def sdf_definition(self, p):
        return Op(OpType.MAX, self.sdf(p), -self.tool(p))
    
class scale(SDF):
    def __init__(self, sdf: SDF, amount: float):
        self.sdf = sdf
        super().__init__(amount=amount)

    def sdf_definition(self, p):
        return self.sdf(p / self.amount) * self.amount



import threading
import queue
import argparse
from pathlib import Path
import importlib.util

import numpy as np
import OpenGL.GL as gl
import glfw
from watchdog.events import FileSystemEvent, FileSystemEventHandler
from watchdog.observers import Observer
from IPython import embed

from topax._utils import compile_shader, rotation_matrix_about_vector, normalize
from topax._builders import Builder
from topax.sdfs import empty
from topax.ops import Const

class SceneHandler:
    QUAD = np.array([
        -1.0, -1.0,
        1.0, -1.0,
        -1.0,  1.0,
        1.0,  1.0
    ], dtype=np.float32)
    VERTEX_SHADER_SOURCE = """
#version 330 core
layout(location = 0) in vec2 aPos;
out vec2 vUV;
void main() {
    vUV = aPos * 0.5 + 0.5;  // map [-1,1] -> [0,1]
    gl_Position = vec4(aPos, 0.0, 1.0);
}
"""
    def __init__(self, window):
        self._window = window
        self._fb_width, self._fb_height = glfw.get_framebuffer_size(window)
        self._program_id = None
        self._shader_source = ""
        self._sdf_repr = None
        self._vao = gl.glGenVertexArrays(1)
        self._vbo = gl.glGenBuffers(1)
        self._camera_position = np.array([0.0, 0.0, 1.0])
        self._camera_up = np.array([0.0, 1.0, 0.0])
        self._looking_at = np.array([0.0, 0.0, 0.0])
        self._fx = 1.0

        self._shader_uniforms = dict(
            i_resolution=None,
            max_steps=None,
            cam_pose=None,
            looking_at=None,
            cam_up=None,
            fx=None,
            stop_epsilon=None,
            tmax=None
        )

        self.update_sdf(empty())

    def __del__(self):
        if self._program_id is not None:
            gl.glDeleteProgram(self._program_id)
        # gl.glDeleteBuffers(self._vbo)
        # gl.glDeleteVertexArrays(self._vao)
    
    def update_sdf(self, sdf):
        sdf_repr = repr(sdf(Const(None, 'p', 'vec3')))
        builder = Builder(sdf)
        if sdf_repr != self._sdf_repr:
            print("recompiling shader!")
            self._sdf_repr = sdf_repr
            shader_code = builder.build()
            input_vars = builder.get_input_vars()

            if self._program_id is not None:
                gl.glDeleteProgram(self._program_id)
            
            vs = compile_shader(SceneHandler.VERTEX_SHADER_SOURCE, gl.GL_VERTEX_SHADER)
            fs = compile_shader(shader_code, gl.GL_FRAGMENT_SHADER)
            self._program_id = gl.glCreateProgram()
            gl.glAttachShader(self._program_id, vs)
            gl.glAttachShader(self._program_id, fs)
            gl.glLinkProgram(self._program_id)
            if not gl.glGetProgramiv(self._program_id, gl.GL_LINK_STATUS):
                raise RuntimeError(gl.glGetProgramInfoLog(self._program_id).decode())
            gl.glDeleteShader(vs)
            gl.glDeleteShader(fs)
            gl.glUseProgram(self._program_id)

            self._shader_uniforms["i_resolution"] = gl.glGetUniformLocation(self._program_id, "_iResolution")
            self._shader_uniforms["max_steps"] = gl.glGetUniformLocation(self._program_id, "_maxSteps")
            self._shader_uniforms["cam_pose"] = gl.glGetUniformLocation(self._program_id, "_camPose")
            self._shader_uniforms["looking_at"] = gl.glGetUniformLocation(self._program_id, "_lookingAt")
            self._shader_uniforms["cam_up"] = gl.glGetUniformLocation(self._program_id, "_camUp")
            self._shader_uniforms["fx"] = gl.glGetUniformLocation(self._program_id, "_fx")
            self._shader_uniforms["stop_epsilon"] = gl.glGetUniformLocation(self._program_id, "_stopEpsilon")
            self._shader_uniforms["tmax"] = gl.glGetUniformLocation(self._program_id, "_tmax")
        else:
            builder.parse_input_vars()
            input_vars = builder.get_input_vars()

        gl.glBindVertexArray(self._vao)
        gl.glBindBuffer(gl.GL_ARRAY_BUFFER, self._vbo)
        gl.glBufferData(gl.GL_ARRAY_BUFFER, SceneHandler.QUAD.nbytes, SceneHandler.QUAD, gl.GL_STATIC_DRAW)
        gl.glEnableVertexAttribArray(0)
        gl.glVertexAttribPointer(0, 2, gl.GL_FLOAT, gl.GL_FALSE, 0, None)

        for input_var, param in input_vars.items():
            loc = gl.glGetUniformLocation(self._program_id, input_var)
            value = param.resolve_value()
            match param.rettype:
                case "vec3":
                    gl.glUniform3f(loc, *value)
                case "vec2":
                    gl.glUniform2f(loc, *value)
                case "float":
                    if hasattr(value, "__iter__"):
                        value = float(value[0])
                    else:
                        value = float(value)
                    gl.glUniform1f(loc, value)
                case _:
                    raise NotImplementedError()

    def rotate_2d(self, dx, dy):
        cam_right = normalize(np.linalg.cross(-self._camera_position, self._camera_up))
        x_rot = rotation_matrix_about_vector(-dx / 300., self._camera_up)
        y_rot = rotation_matrix_about_vector(-dy / 300., cam_right)
        self._camera_position = x_rot @ self._camera_position
        self._camera_position = y_rot @ self._camera_position
        self._camera_up = y_rot @ self._camera_up

    def zoom(self, delta):
        factor = (1 + delta * 0.008)
        self._fx *= factor
        self._camera_position *= factor
        

    def draw_scene(self):
        """
        This function is responsible for drawing all parts of the scene. It will take in the 
        """
        gl.glViewport(0, 0, self._fb_width, self._fb_height)
        gl.glClearColor(0.2, 0.2, 0.2, 1.0)
        gl.glClear(gl.GL_COLOR_BUFFER_BIT | gl.GL_DEPTH_BUFFER_BIT)

        gl.glEnable(gl.GL_BLEND)
        gl.glBlendFunc(gl.GL_SRC_ALPHA, gl.GL_ONE_MINUS_SRC_ALPHA)

        gl.glUniform2f(self._shader_uniforms["i_resolution"], self._fb_width, self._fb_height)
        gl.glUniform1ui(self._shader_uniforms["max_steps"], 256)
        gl.glUniform3f(self._shader_uniforms["cam_pose"], *self._camera_position)
        gl.glUniform3f(self._shader_uniforms["looking_at"], *self._looking_at)
        gl.glUniform3f(self._shader_uniforms["cam_up"], *self._camera_up)
        gl.glUniform1f(self._shader_uniforms["fx"], self._fx)
        gl.glUniform1f(self._shader_uniforms["stop_epsilon"], 0.0001)
        gl.glUniform1f(self._shader_uniforms["tmax"], 1000.0)

        gl.glBindVertexArray(self._vao)
        gl.glDrawArrays(gl.GL_TRIANGLE_STRIP, 0, 4)

        glfw.swap_buffers(self._window)

class FileEventHandler(FileSystemEventHandler):
    def __init__(self, callback):
        self._on_modified = callback

    def on_modified(self, event: FileSystemEvent) -> None:
        self._on_modified(event)

class ProjectInterface:
    def __init__(self, root_path, sdf_queue: queue.Queue):
        self._root_path = root_path
        self._target = None
        self._sdf_queue = sdf_queue
        self._thread = threading.Thread(target=self.repl_worker, daemon=True)
        self._event_handler = FileEventHandler(self._file_change_event)
        self._observer = Observer()
        self._observer.schedule(self._event_handler, self._root_path, recursive=True)
        self._observer.start()
        self._thread.start()

    def repl_worker(self):
        banner = "CAX REPL started. Use shared_state/command_queue to talk to renderer."
        embed(header=banner, banner1="", colors="neutral", user_ns=dict(target=self.set_target_file))

    def set_target_file(self, path):
        path = Path(self._root_path, path)
        print("watching ", path)
        if not path.exists():
            print("targeted file doesn't exist!")
            return
        self._target = path
        self._sdf_queue.put(self._target)
        glfw.post_empty_event()

    def _file_change_event(self, event):
        if self._target is None: return
        try:
            if Path(event.src_path).samefile(self._target):
                self._sdf_queue.put(self._target)
                glfw.post_empty_event()
        except FileNotFoundError as e:
            pass

def main():
    # Parse argments
    parser = argparse.ArgumentParser()
    parser.add_argument("dir", help="top level directory of CAD project")
    args = parser.parse_args()
    project_dir = Path(args.dir)

    if not project_dir.is_dir():
        raise FileNotFoundError(f"Can't find project dir {project_dir}")

    # Initialize glfw window
    if not glfw.init():
        raise RuntimeError("Failed to initialize GLFW")
    
    glfw.window_hint(glfw.CONTEXT_VERSION_MAJOR, 3)
    glfw.window_hint(glfw.CONTEXT_VERSION_MINOR, 3)
    glfw.window_hint(glfw.OPENGL_FORWARD_COMPAT, gl.GL_TRUE)
    glfw.window_hint(glfw.OPENGL_PROFILE, glfw.OPENGL_CORE_PROFILE)

    window = glfw.create_window(800, 600, "TOPAX", None, None)
    if not window:
        glfw.terminate()
        raise RuntimeError("Failed to create GLFW window")

    glfw.make_context_current(window)

    # Initialize scene handler
    scene = SceneHandler(window)

    # Initialize callbacks
    dragging = False
    last_pos_x, last_pos_y = 0, 0
    def mouse_button_callback(win, button, action, mods):
        nonlocal dragging, last_pos_x, last_pos_y
        if button == glfw.MOUSE_BUTTON_LEFT:
            dragging = (action == glfw.PRESS)
            last_pos_x, last_pos_y = glfw.get_cursor_pos(window)
            if not dragging:
                scene.draw_scene()

    def scroll_callback(win, xoffset, yoffset):
        scene.zoom(yoffset)
        scene.draw_scene()

    glfw.set_mouse_button_callback(window, mouse_button_callback)
    glfw.set_scroll_callback(window, scroll_callback)

    # Setup command message queue and project interface
    sdf_queue = queue.Queue()
    project_interface = ProjectInterface(project_dir, sdf_queue)

    scene.draw_scene()

    # Main application loop
    while not glfw.window_should_close(window):
        if dragging:
            x, y = glfw.get_cursor_pos(window)
            dx = x - last_pos_x
            dy = y - last_pos_y
            last_pos_x = x
            last_pos_y = y

            if dx != 0 or dy != 0:
                scene.rotate_2d(dx, dy)
                scene.draw_scene()

        while not sdf_queue.empty():
            print("sdf queue item")
            new_sdf = sdf_queue.get()
            # sys.modules['cax.sdfs'] = cax.sdfs
            spec = importlib.util.spec_from_file_location("_external_script", new_sdf)
            module = importlib.util.module_from_spec(spec)
            spec.loader.exec_module(module)
            scene.update_sdf(module.make_part())
            scene.draw_scene()
            
        glfw.wait_events()

    # Clean up after app closes
    glfw.terminate()
    del scene

if __name__ == "__main__":
    main()



import jinja2

from topax._utils import make_shader_const
from topax.ops import Const, Op, OpType

class Builder:
    template = jinja2.Environment(loader=jinja2.PackageLoader('topax')).get_template('shader.glsl.j2')
    def __init__(self, sdf):
        self.sdf = sdf
        self.p = Const(None, 'p', 'vec3')
        self.optree = self.sdf(self.p)
        self.input_vars = {} # keyed by Const objects, value is varname
        self.input_vars_types = {} # keyed by input var name, value is type
        self.local_vars = {} # keyed by name of var, element is type
        self.local_vars_unused = {} # keyed by type, each element is list
        self.lines = []

    def _allocate_var(self, vartype):
        if vartype not in self.local_vars_unused:
            self.local_vars_unused[vartype] = []
        if len(self.local_vars_unused[vartype]) > 0:
            return self.local_vars_unused[vartype].pop()
        else:
            new_varname = f"local_var{len(self.local_vars.keys())}"
            self.local_vars[new_varname] = vartype
            self.lines.append(f"{vartype} {new_varname};")
            return new_varname

    def _release_var(self, varname):
        if varname is not None:
            if varname in self.input_vars_types: return
            assert varname in self.local_vars
            vartype = self.local_vars[varname]
            self.local_vars_unused[vartype].append(varname)

    def _parse_input_vars(self, optree: Op | Const):
        if isinstance(optree, Const):
            if optree in self.input_vars: return
            if optree.sdf is None:
                # self.input_vars[optree] = optree.param
                # self.input_vars_types[optree.param] = optree.rettype
                pass
            else:
                new_name = f"sdfin_var{len(self.input_vars.keys())}"
                self.input_vars[optree] = new_name
                self.input_vars_types[new_name] = optree.rettype
        else:
            self._parse_input_vars(optree.lhs)
            if optree.rhs is not None: self._parse_input_vars(optree.rhs)

    def parse_input_vars(self):
        self._parse_input_vars(self.optree)
    
    def get_input_vars(self):
        return {value: key for key, value in self.input_vars.items()}

    def _parse_ops(self, optree: Op | Const, out_varname: str):
        # deal with case for empty or constant SDF
        if isinstance(optree, Const):
            assert optree.sdf is None
            self.lines.append(f"{out_varname} = {optree.param};")
            return
        
        # retrieve shader var name for lhs
        lhs_is_external_param = False
        if isinstance(optree.lhs, Const):
            if optree.lhs.sdf is None:
                lhs_is_external_param = True
                if isinstance(optree.lhs.param, str): lhs_varname = optree.lhs.param
                else: lhs_varname = make_shader_const(optree.lhs.param, optree.lhs.rettype)
            else:
                lhs_varname = self.input_vars[optree.lhs]
        else:
            lhs_varname = self._allocate_var(optree.lhs.rettype)
            self._parse_ops(optree.lhs, lhs_varname)
        
        # retrieve shader var name for rhs
        rhs_is_external_param = False
        if optree.rhs is not None:
            if isinstance(optree.rhs, Const):
                if optree.rhs.sdf is None:
                    rhs_is_external_param = True
                    if isinstance(optree.rhs.param, str): rhs_varname = optree.rhs.param
                    else: rhs_varname = make_shader_const(optree.rhs.param, optree.rhs.rettype)
                else:
                    rhs_varname = self.input_vars[optree.rhs]
            else:
                rhs_varname = self._allocate_var(optree.rhs.rettype)
                self._parse_ops(optree.rhs, rhs_varname)
        else:
            rhs_varname = None

        match optree.opcode:
            case OpType.ADD: self.lines.append(f"{out_varname} = {lhs_varname} + {rhs_varname};")
            case OpType.SUB: self.lines.append(f"{out_varname} = {lhs_varname} - {rhs_varname};")
            case OpType.MUL: self.lines.append(f"{out_varname} = {lhs_varname} * {rhs_varname};")
            case OpType.DIV: self.lines.append(f"{out_varname} = {lhs_varname} / {rhs_varname};")
            case OpType.LEN: self.lines.append(f"{out_varname} = length({lhs_varname});")
            case OpType.NORM: self.lines.append(f"{out_varname} = normal({lhs_varname});")
            case OpType.MIN: self.lines.append(f"{out_varname} = min({lhs_varname}, {rhs_varname});")
            case OpType.MAX: self.lines.append(f"{out_varname} = max({lhs_varname}, {rhs_varname});")
            case OpType.NEG: self.lines.append(f"{out_varname} = -{lhs_varname};")
            case OpType.ABS: self.lines.append(f"{out_varname} = abs({lhs_varname});")
            case OpType.X: self.lines.append(f"{out_varname} = {lhs_varname}.x;")
            case OpType.Y: self.lines.append(f"{out_varname} = {lhs_varname}.y;")
            case OpType.Z: self.lines.append(f"{out_varname} = {lhs_varname}.z;")
            case OpType.XY: self.lines.append(f"{out_varname} = {lhs_varname}.xy;")
            case OpType.XZ: self.lines.append(f"{out_varname} = {lhs_varname}.xz;")
            case OpType.YZ: self.lines.append(f"{out_varname} = {lhs_varname}.yz;")
            case _: raise NotImplementedError(f"parsing for opcode {optree.opcode} not supported")

        if not lhs_is_external_param: self._release_var(lhs_varname)
        if not rhs_is_external_param: self._release_var(rhs_varname)

    def build(self):
        self.parse_input_vars()
        self._parse_ops(self.optree, 'd')
        
        return Builder.template.render(
            global_inputs=[dict(name=v, type=self.input_vars_types[v]) for v in self.input_vars_types if v != 'p'],
            lines=self.lines
        )
